-- SPDX-FileCopyrightText: 2021 Laouen Fernet
--
-- SPDX-License-Identifier: GPL-3.0-or-later

module Lib where
import qualified Data.List                     as L
import qualified Data.Map                      as M
import           Data.Maybe                     ( catMaybes
                                                , fromMaybe
                                                , isJust
                                                , mapMaybe
                                                , maybeToList
                                                )

-- | A function symbol is represented as a string.
type Function = String
-- | A variable name is represented as a string.
type Variable = String

-- | A @Term@ is either a @Variable@ or a @Function@ applied to a list of
-- subterms. A constant is a @Function@ applied to the empty list (no subterms).
data Term = Var Variable
          | Fun Function [Term]
          deriving (Eq, Ord)

instance Show Term where
  show (Var x   ) = x
  show (Fun f []) = f
  show (Fun f ts) = f ++ "(" ++ L.intercalate "," (L.map show ts) ++ ")"

-- | A recipe is a @Function@ applied to subrecipes. It should not be a
-- @Variable@, since this is not enforced by this definition it will be checked
-- when relevant in functions using recipes.
type Recipe = Term
-- | A @Substitution@ maps @Variable@ to @Term@.
type Substitution = M.Map Variable Term
-- | A @Frame@ maps @Recipe@ (called labels) to @Term@. It is used to represent
-- the knowledge of the intruder.
type Frame = M.Map Recipe Term

-- | Convert a @Substitution@ to a pretty string.
showSub :: M.Map Variable Term -> String
showSub sigma =
  "{"
    ++ L.intercalate
         ", "
         (L.map (\(x, t) -> x ++ " -> " ++ show t) (M.assocs sigma))
    ++ "}"

-- | Convert a @Frame@ to a pretty string.
showFrame :: M.Map Recipe Term -> String
showFrame frame =
  "{|"
    ++ L.intercalate
         ", "
         (L.map (\(l, t) -> show l ++ " -> " ++ show t) (M.assocs frame))
    ++ "|}"
--
-- | Convert a list of equalities to a "pretty" string.
-- The logical formula is the conjunction of all lines.
showEqs :: [(Term, Term)] -> String
showEqs eqs = "Equalities:\n"
  ++ L.intercalate "\n" (L.map (\(s, t) -> show s ++ " = " ++ show t) eqs)

-- | Convert a list of inequalities to a "pretty" string.
-- The logical formula is the conjunction of all lines.
showIneqs :: [[(Term, Term)]] -> String
showIneqs ineqs = "Inequalities:\n"
  ++ L.intercalate "\n" (L.map showIneq ineqs)
 where
  showIneq = L.intercalate " | "
    . L.map (\(s, t) -> "(" ++ show s ++ " != " ++ show t ++ ")")

-- | Substitute all @Variable@ in a @Term@ according to @Substitution@.
substitute :: Substitution -> Term -> Term
substitute sigma (Var x   ) = fromMaybe (Var x) $ M.lookup x sigma
substitute sigma (Fun f ts) = Fun f $ L.map (substitute sigma) ts

-- | Compute the list of @Variable@ occuring in a @Term@.
vars :: Term -> [Variable]
vars (Var x   ) = [x]
vars (Fun f ts) = concatMap vars ts

-- | @True@ if and only if @Term@ is ground.
ground :: Term -> Bool
ground = null . vars

-- | Apply a @Substitution@ over a list of equalitites.
substituteList :: Substitution -> [(Term, Term)] -> [(Term, Term)]
substituteList sigma =
  L.map (\(x, y) -> (substitute sigma x, substitute sigma y))

-- | Compute a most general unifier of the equalities given. The result is a
-- @Maybe Substitution@, which is @Nothing@ if there is no unifier.
unify :: [(Term, Term)] -> Substitution -> Maybe Substitution
unify []            sigma = Just sigma
unify ((s, t) : ts) sigma = if s == t
  then unify ts sigma -- Delete
  else case (s, t) of
    (Fun f us, Fun g vs) -> if f == g
      then unify (zip us vs ++ ts) sigma -- Decompose
      else Nothing -- Clash
    (Fun f us, Var x) -> unify ((t, s) : ts) sigma -- Orient
    (Var x   , _    ) -> if x `elem` vars t
      then Nothing -- Occurs check
      else
        let sigma' = M.insert x t sigma
        in  unify (substituteList sigma' ts) sigma' -- Eliminate

-- | Compute a most general unifier of the equality between two @Term@.
unifyEq :: Term -> Term -> Maybe Substitution
unifyEq s t = unify [(s, t)] M.empty

-- | Names of public functions.
publistFun =
  [ "pub"
  , "priv"
  , "crypt"
  , "dcrypt"
  , "vcrypt"
  , "sign"
  , "retrieve"
  , "vsig"
  , "scrypt"
  , "dscrypt"
  , "vscrypt"
  , "pair"
  , "proj1"
  , "proj2"
  , "vpair"
  , "h"
  ]

-- | A @Function@ is public if it is in @publistFun@.
public :: String -> Bool
public f = f `elem` publistFun

-- | Names of variables with public range.
publistVar = ["v1", "v2", "v3"]

-- | A @Variable@ has a public range if it is in @publistVar@.
pubvar :: String -> Bool
pubvar x = x `elem` publistVar

-- | Compute the cartesian product of a list of lists.
cartProd :: [[a]] -> [[a]]
cartProd []         = [[]]
cartProd (xs : xss) = [ x : ys | let yss = cartProd xss, x <- xs, ys <- yss ]

-- | Convert a @Substitution@ to a list of equalities.
subToEqs :: Substitution -> [(Term, Term)]
subToEqs = L.map (\(x, t) -> (Var x, t)) . M.toList

-- | Compute a list of @Recipe@ to compose a ground @Term@ from a concrete
-- @Frame@.
-- A @Variable@ cannot be composed, as the @Term@ should be ground. The @Term@
-- can be composed if there is a label mapping to it, or if the top level
-- @Function@ is public and all subterms can be composed. The different
-- combinations of recipes for the subterms are computed as a cartesian product.
compose :: Frame -> Term -> [Recipe]
compose concr (Var x) =
  error
    $  "Trying to compose variable "
    ++ x
    ++ " but only ground terms can be composed."
compose concr t@(Fun f ts) =
  let rs  = M.keys $ M.filter (t ==) concr
      us  = L.map (compose concr) ts
      rs' = if public f then rs ++ L.map (Fun f) (cartProd us) else rs
  in  L.nub rs'

-- | Compute a list of pairs (@Recipe@, @Substitution@) to compose a @Term@ from
-- a structural @Frame@.
-- The @Term@ can be composed by unifying with the @Term@ that are in the
-- @Frame@. For a @Variable@ with public range, the value can be directly read
-- from the @Substitution@, and the ground @Term@ it maps to is also a @Recipe@.
-- For a public @Function@, all combinations of composition for the subterms are
-- also computed, as in @compose@.
composeUnder :: Substitution -> Frame -> Term -> [(Recipe, Substitution)]
composeUnder theta struct t@(Var x) =
  let rus =
        [ (l, sigma)
        | l     <- M.keys struct
        , sigma <- maybeToList $ do
          s <- M.lookup l struct
          unifyEq t s
        ]
      rus' = if pubvar x
        then case M.lookup x theta of
          Nothing -> rus
          Just c  -> (c, M.singleton x c) : rus
        else rus
  in  L.nub rus'
composeUnder theta struct t@(Fun f ts) =
  let
    rus =
      [ (l, sigma)
      | l     <- M.keys struct
      , sigma <- maybeToList $ do
        s <- M.lookup l struct
        unifyEq t s
      ]
    us   = L.map (composeUnder theta struct) ts
    rus' = if public f
      then
        rus
          ++ [ (Fun f $ L.map fst vs, sigma)
             | vs <- cartProd us
             , let unifier = unify (L.concatMap (subToEqs . snd) vs) M.empty
             , sigma <- maybeToList unifier
             ]
      else rus
  in
    L.nub rus'

-- | Apply a @Recipe@ to a @Frame@.
-- Substitute the labels in the @Recipe@ by the @Term@ they map to in the
-- @Frame@. @Variable@ are not allowed to occur in @Recipe@. This is the
-- difference between @Recipe@ and @Term@, which is not enforced by the type
-- definitions.
cook :: Frame -> Recipe -> Term
cook frame (  Var x   ) = error $ "Cannot apply recipe, contains variable " ++ x
cook frame r@(Fun f ts) = case M.lookup r frame of
  Nothing -> Fun f $ L.map (cook frame) ts
  Just t  -> t

-- | Compute a list of pairs from a list of lists.
-- Each list is interpreted as an equivalence class, the pairs returned avoid
-- redundancy by relying on symmetry and transitivity. Singletons are ignored,
-- as they cannot generate pairs.
pairsEcs :: Eq a => [[a]] -> [(a, a)]
pairsEcs ecs =
  [ (x, y) | ec <- ecs, not $ null ec, let x = head ec, y <- tail ec ]

-- | Apply a @Substitution@ to all @Term@ in a structural @Frame@.
-- @Term@ in the concrete @Frame@ must be ground.
getConcr :: Substitution -> Frame -> Frame
getConcr theta struct =
  let concr = M.map (substitute theta) struct
  in  if not . L.all ground $ M.elems concr
        then error "The substitution does not map all variables in the frame."
        else concr

-- | Compute the list of @Term@ required (keys) and the list of derivable @Term@
-- resulting from analysing a @Term@.
-- Each derivable @Term@ is paired with a @Function@, the destructor to be used
-- as part of the @Recipe@ to compose the @Term@.
ana :: Term -> ([Term], [(Function, Term)])
ana (Fun "scrypt" [k, t]) = ([k], [("dscrypt", t)])
ana (Fun "crypt" [Fun "pub" [s], t]) = ([Fun "priv" [s]], [("dcrypt", t)])
ana (Fun "sign" [Fun "priv" [s], t]) = ([], [("retrieve", t)])
ana (Fun "pair" [t1, t2]) = ([], [("proj1", t1), ("proj2", t2)])
ana _ = ([], [])

-- | Analyse recursively pairs (@Recipe@, @Term@) at the structural level,
-- assuming static equivalence to a corresponding concrete frame.
-- Saturation is achieved when no @Term@ can be further analysed. The pairs
-- correspond to elements in a structural frame. The @Substitution@ maps
-- @Variable@ to their concrete value. If the analysis is successful, @Recipe@
-- for the new @Term@ use the @Recipe@ of keys in the concrete frame. If the
-- analysis fails, the @Substitution@ paired with @Recipe@ for keys in the
-- structural frame are not consistent with the static equivalence, they are
-- called exceptions.
analyseRec
  :: Substitution            -- ^ The concrete instantiation of variables
  -> [(Recipe, Term)]        -- ^ The list of pairs to analyse
  -> [(Recipe, Term)]        -- ^ The list of pairs that could be analysed again later
  -> [(Recipe, Term)]        -- ^ The list of completely analysed pairs
  -> [Substitution]          -- ^ The list of exceptions
  -> (Frame, [Substitution]) -- ^ The analysed frame and the list of exceptions
analyseRec theta [] hold done exceptions =
  (M.fromList (hold ++ done), L.nub exceptions)
analyseRec theta ((l, t) : lts) hold done exceptions =
  let
    (ks, fts) = ana t
    ftsNew    = L.filter (\(_, s) -> s `notElem` L.map snd done) fts
  in
    if null ftsNew
      then analyseRec theta lts hold ((l, t) : done) exceptions
      else
        let
          struct = M.fromList $ (l, t) : lts ++ hold ++ done
          concr  = getConcr theta struct
          krs    = zip ks $ L.map (compose concr . substitute theta) ks
        in
          if L.any (null . snd) krs
            then
              let exceptionsNew =
                    [ sigma
                    | us <- cartProd $ L.map (composeUnder theta struct) ks
                    , let unifier =
                            unify (L.concatMap (subToEqs . snd) us) M.empty
                    , sigma <- maybeToList unifier
                    ]
              in  analyseRec theta
                             lts
                             ((l, t) : hold)
                             done
                             (exceptions ++ exceptionsNew)
            else
              let
                ltsNew =
                  [ (Fun f $ rs ++ [l], s)
                  | let rs = L.map (head . snd) krs
                  , (f, s) <- ftsNew
                  ]
                  ++ [ (r, k)
                     | (k, rs) <- krs
                     , let r = head rs
                     , r `notElem` M.keys struct
                     ]
              in  analyseRec theta
                             (ltsNew ++ lts ++ hold)
                             []
                             ((l, t) : done)
                             exceptions

-- | Analyse a structural @Frame@, with a @Substitution@ mapping all @Variable@
-- occurring in the @Frame@ to ground @Term@.
-- Call the recursive analysis function with all elements in the @Frame@ as new
-- pairs to be analysed.
analyse
  :: Substitution            -- ^ The concrete instantiation of variables
  -> Frame                   -- ^ The structural frame
  -> (Frame, [Substitution]) -- ^ The analysed frame with the list of exceptions
analyse theta struct = analyseRec theta (M.toList struct) [] [] []

analyseRecBis
  :: Substitution            -- ^ The concrete instantiation of variables
  -> [(Recipe, Term)]        -- ^ The list of pairs to analyse
  -> [(Recipe, Term)]        -- ^ The list of pairs that could be analysed again later
  -> [(Recipe, Term)]        -- ^ The list of completely analysed pairs
  -> [Substitution]          -- ^ The list of exceptions
  -> (Frame, [Substitution]) -- ^ The analysed frame and the list of exceptions
analyseRecBis theta [] hold done exceptions =
  (M.fromList (hold ++ done), L.nub exceptions)
analyseRecBis theta ((l, t) : lts) hold done exceptions =
  let
    (ks, fts) = ana t
    ftsNew    = L.filter (\(_, s) -> s `notElem` L.map snd done) fts
  in
    if null ftsNew
      then analyseRecBis theta lts hold ((l, t) : done) exceptions
      else
        let
          struct = M.fromList $ (l, t) : lts ++ hold ++ done
          concr  = getConcr theta struct
          krs    = zip ks $ L.map (compose concr . substitute theta) ks
        in
          if L.any (null . snd) krs
            then
              let exceptionsNew =
                    [ sigma
                    | us <- cartProd $ L.map (composeUnder theta struct) ks
                    , let unifier =
                            unify (L.concatMap (subToEqs . snd) us) M.empty
                    , sigma <- maybeToList unifier
                    ]
              in  analyseRecBis theta
                             lts
                             ((l, t) : hold)
                             done
                             (exceptions ++ exceptionsNew)
            else
              let
                ltsNew =
                  [ (Fun f $ rs ++ [l], s)
                  | let rs = L.map (head . snd) krs
                  , (f, s) <- ftsNew
                  ]
              in  analyseRecBis theta
                             (ltsNew ++ lts ++ hold)
                             []
                             ((l, t) : done)
                             exceptions

-- | Analyse a structural @Frame@, with a @Substitution@ mapping all @Variable@
-- occurring in the @Frame@ to ground @Term@.
-- Call the recursive analysis function with all elements in the @Frame@ as new
-- pairs to be analysed.
analyseBis
  :: Substitution            -- ^ The concrete instantiation of variables
  -> Frame                   -- ^ The structural frame
  -> (Frame, [Substitution]) -- ^ The analysed frame with the list of exceptions
analyseBis theta struct = analyseRecBis theta (M.toList struct) [] [] []


-- | Analyse a list of @Frame@ with a @Substitution@.
-- The different @Frame@ are possibilities considered by the intruder. Every
-- possibility is analysed separately, resulting in a number of analysed @Frame@
-- paired with exceptions.
analyseMulti
  :: Substitution              -- ^ The concrete instantiation of variables
  -> [Frame]                   -- ^ The list of possible structural frames
  -> [(Frame, [Substitution])] -- ^ The list of analysed frames with exceptions
analyseMulti theta = L.map $ analyse theta

-- | Compute a list of equalities and a list of inequalities between @Variable@
-- occurring in a structural @Frame@, based on static equivalence with a
-- concrete @Frame@.
-- The frames are assumed to have been already analysed. Equalities between
-- @Variable@ are found when there are several recipes to compose a term in the
-- concrete frame. Inequalities are found when there are pairs of recipes
-- distinguishing the two frames. The unifiers such that terms are equal in the
-- structural frame but not in the concrete frame are not consistent with the
-- static equivalence, they are exceptions.
relations
  :: Substitution                       -- ^ The concrete instantiation of variables
  -> Frame                              -- ^ The structural frame
  -> Frame                              -- ^ The concrete frame
  -> [Substitution]                     -- ^ The list of exceptions found so far
  -> ([(Term, Term)], [[(Term, Term)]]) -- ^ The list of equalities and the list of inequalities
relations theta struct concr exceptions =
  let pairs = pairsEcs . L.map (compose concr) $ M.elems concr
      eqs   = L.map (\(r1, r2) -> (cook struct r1, cook struct r2)) pairs
      ineqs = L.map subToEqs $ exceptions ++ L.nub
        [ sigma
        | (l, t    ) <- M.assocs struct
        , (r, sigma) <- composeUnder theta struct t
        , l /= r
        , sigma `notElem` exceptions
        , cook concr l /= cook concr r
        ]
      unifier = unify eqs M.empty
  in  case unifier of
        Nothing    -> error $ "There is no unifier! " ++ showEqs eqs
        Just sigma -> (subToEqs sigma, ineqs)

-- | Find relations between the variables occurring in a structural @Frame@,
-- based on static equivalence with the concrete @Frame@ found with the
-- @Substitution@.
-- The @Frame@ is first analysed. Then @relations@ is called on the result of the analysis.
findRelations
  :: Substitution                       -- ^ The concrete instantiation of variables
  -> Frame                              -- ^ The structural frame
  -> ([(Term, Term)], [[(Term, Term)]]) -- ^ The list of equalities and the list of inequalities
findRelations theta struct =
  let (structAna, exceptions) = analyse theta struct
      concrAna                = getConcr theta structAna
  in  relations theta structAna concrAna exceptions

-- | Find relations between variables for a list of possible structural @Frame@,
-- using the same @Substitution@ for each of the possibility.
-- When the intruder considers several possibilities, each frame is analysed
-- and relations are found for each possibility. The result is a list of
-- analysed @Frame@, each with a list of equalities and a list of inequalities.
findRelationsMulti
  :: Substitution                                  -- ^ The concrete instantiation of variables
  -> [Frame]                                       -- ^ The list of possible structural frames
  -> [(Frame, ([(Term, Term)], [[(Term, Term)]]))] -- ^ The list of analysed frames with equalities and inequalities
findRelationsMulti theta [] =
  error "The list of possibilities is empty, at least one frame is required."
findRelationsMulti theta structs@(struct1 : structis) =
  let structAnaExs = analyseMulti theta structs
      struct1ana   = fst $ head structAnaExs
      concr1ana    = getConcr theta struct1ana
  in  [ (structAna, relations theta structAna concr1ana exceptions)
      | (structAna, exceptions) <- structAnaExs
      , let concrAna = getConcr theta structAna
      , areStatEq concr1ana concrAna
      ]

-- | Show the relations found, for a @Substitution@, between the variables
-- occurring in a structural @Frame@.
-- @findRelations@ is called and the result is "pretty" printed.
showRelations :: Substitution -> Frame -> IO ()
showRelations theta struct =
  let (eqs, ineqs) = findRelations theta struct
  in  putStrLn
        $  "theta = "
        ++ showSub theta
        ++ "\nstruct = "
        ++ showFrame struct
        ++ "\n"
        ++ showEqs eqs
        ++ "\n"
        ++ showIneqs ineqs

-- | Show the relations found, for a @Substitution@, between the variables
-- occurring in a list of possible structural @Frame@.
-- @findRelationsMulti@ is called and the result is "pretty" printed, with a
-- separation between each possibility.
showRelationsMulti :: Substitution -> [Frame] -> IO ()
showRelationsMulti theta structs =
  putStrLn $ "theta = " ++ showSub theta ++ L.concatMap
    showPoss
    (findRelationsMulti theta structs)
 where
  showPoss ((struct, (eqs, ineqs))) =
    "\n\nHypothesis: struct = "
      ++ showFrame struct
      ++ "\n"
      ++ showEqs eqs
      ++ "\n"
      ++ showIneqs ineqs

-- | @True@ if and only if the two concrete @Frame@ are statically equivalent.
-- The frames are assumed to be ground and to have been already analysed.
areStatEq :: Frame -> Frame -> Bool
areStatEq frame1 frame2 =
  let pairs1 = pairsEcs . L.map (compose frame1) $ M.elems frame1
      pairs2 = pairsEcs . L.map (compose frame2) $ M.elems frame2
  in  L.and
        $  L.map (\(r1, r2) -> cook frame2 r1 == cook frame2 r2) pairs1

        ++ L.map (\(r1, r2) -> cook frame1 r1 == cook frame1 r2) pairs2

test1struct = M.fromList
  [(Fun "m0" [], Fun "scrypt" [Var "x", Var "y"]), (Fun "m1" [], Var "z")]
test1theta =
  M.fromList [("x", Fun "k1" []), ("y", Fun "msg" []), ("z", Fun "k1" [])]
test1 = showRelations test1theta test1struct
test2theta =
  M.fromList [("x", Fun "k1" []), ("y", Fun "msg" []), ("z", Fun "k2" [])]
test2 = showRelations test2theta test1struct
test3struct = M.fromList
  [ (Fun "m0" [], Fun "h" [Var "v1"])
  , (Fun "m1" [], Fun "h" [Var "v2"])
  , (Fun "m2" [], Fun "h" [Var "v3"])
  ]
test3theta =
  M.fromList [("v1", Fun "0" []), ("v2", Fun "1" []), ("v3", Fun "0" [])]
test3 = showRelations test3theta test3struct
test4theta =
  M.fromList [("v1", Fun "0" []), ("v2", Fun "1" []), ("v3", Fun "2" [])]
test4 = showRelations test4theta test3struct

test5struct1 = M.fromList
  [ (Fun "m0" [], Fun "scrypt" [Fun "k1" [], Var "x"])
  , (Fun "m1" [], Fun "k1" [])
  , (Fun "m2" [], Fun "pair" [Var "y", Var "z"])
  , (Fun "m3" [], Fun "k2" [])
  ]
test5struct2 = M.fromList
  [ (Fun "m0" [], Fun "scrypt" [Fun "k1" [], Var "x"])
  , (Fun "m1" [], Fun "k2" [])
  , (Fun "m2" [], Fun "pair" [Var "y", Var "z"])
  , (Fun "m3" [], Fun "k1" [])
  ]
test5theta = M.fromList
  [("x", Fun "success" []), ("y", Fun "hello" []), ("z", Fun "world" [])]
test5 = showRelationsMulti test5theta [test5struct1, test5struct2]

test6struct1 = M.fromList
  [ (Fun "m0" [], Fun "scrypt" [Fun "k1" [], Var "x"])
  , (Fun "m1" [], Var "u")
  , (Fun "m2" [], Fun "pair" [Var "y", Var "z"])
  , (Fun "m3" [], Fun "h" [Var "u"])
  ]
test6struct2 = M.fromList
  [ (Fun "m0" [], Fun "scrypt" [Fun "k1" [], Var "x"])
  , (Fun "m1" [], Var "u")
  , (Fun "m2" [], Fun "pair" [Var "y", Var "z"])
  , (Fun "m3" [], Fun "h" [Var "v"])
  ]
test6theta = M.fromList
  [ ("x", Fun "success" [])
  , ("y", Fun "hello" [])
  , ("z", Fun "world" [])
  , ("u", Fun "k1" [])
  , ("v", Fun "k2" [])
  ]
test6 = showRelationsMulti test6theta [test6struct1, test6struct2]
